from __future__ import annotations

import inspect
import logging
import typing
from re import match
from typing import Optional, TypeVar, Union, get_type_hints

import returns
from returns.maybe import Maybe, Nothing, Some, maybe
from returns.pipeline import pipe
from returns.result import Failure, Result, Success

_log = logging.getLogger(__name__)

S = TypeVar('S')
T = TypeVar('T')


class Unresolvable(Exception):

    def __init__(self, type: T):
        super().__init__(f"Couldn't resolve type: {type} in container.")


class ResolutionError(Exception):
    pass


# Descriptor pattern from
class Frozen:
    __slots__ = ('private_name', )

    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype: T = None) -> T:
        value = getattr(obj, self.private_name)    # type:ignore
        return value

    def __set__(self, obj: T, value):
        if hasattr(obj, self.private_name):    # type:ignore
            msg = f'Attribute `{self.private_name[1:]}` is immutable!'
            raise TypeError(msg)

        setattr(obj, self.private_name, value)    # type:ignore


class Container:
    containers: dict[Container] = {}
    name: Frozen = Frozen()

    def start_services(self):
        pass

    class Resolvable:

        @property
        def type(self) -> T:
            raise NotImplementedError("type property must be implemented")

        @property
        def instance(self) -> S:
            raise NotImplementedError("instance property must be implemented")

        def resolve(self) -> Result[T, None]:
            raise NotImplementedError("resolve method must be implemented")

    # Container __init__
    def __init__(self, name: str = None) -> None:
        if name in self.containers:
            raise AttributeError(f"Container: {name} already exists and can't be modified.")
        self.name = name
        Container.containers[name] = self
        self._resolvable: dict[T, S] = {}
        self._resolvable_lists: dict[T, Container.ResolvableList] = {}

    @staticmethod
    def destroy_all():
        for k, v in Container.containers.items():
            v._resolvable.clear()
        Container.containers.clear()

    @staticmethod
    def is_optional_type(field) -> bool:
        return typing.get_origin(field) is Union and \
            type(None) in typing.get_args(field)

    def _resolve_optional_argument(self, optional_field: T) -> Result[T, None]:

        for option in typing.get_args(optional_field):
            if option is not None:
                return option
            return next(filter(lambda l: l is not None, typing.get_args(optional_field)))

    def _resolve_argument(self, field: T) -> Result[T, None]:

        resolved = None
        if field in self._resolvable:
            container = self._resolvable[field]
            if isinstance(container, Container.Resolvable):
                resolved = container.resolve()
            else:
                resolved = Success(container)
        else:
            for k, v in self._resolvable.items():
                if isinstance(v, Container.ResolvableList):
                    if v.contains(field):
                        resolved = v.retrieve(field)
                        break
                    elif found_subclass := v.find_subclass(field):
                        return v.retrieve(found_subclass)

            if not resolved:
                if self.is_optional_type(field):
                    resolved = Success(None)
                else:
                    resolved = Failure(f"Couldn't resolve {field}")

            # resolved = Failure(f"Couldn't resolve {field}")
            # if isinstance(self._resolvable[field], Container.Singleton) or isinstance(
            #         self._resolvable[field], Container.Transient):
            #     _log.debug(f"For k: {k} using: {self._resolvable[field]._value} to resolve (.value) for.")
            #     self._kwargs[k] = self._resolvable[field].value
            # else:
            #     _log.debug(f"For k: {k} using: {self._resolvable[field]} to resolve (.value) for.")
            #     self._kwargs[k] = self._resolvable[field]
        return resolved

    def _resolve(self, field: T, **kwargs) -> Result[T, None]:

        found = self._resolvable.get(field)

        if not found:
            for k, v in self._resolvable_lists.items():
                if v.contains(field):
                    return v.retrieve(field)

                found_subclass = v.find_subclass(field)
                if found_subclass:
                    return v.retrieve(found_subclass)

        # If already found
        if found:
            if isinstance(found, Container.Resolvable):
                match found.resolve(**kwargs):
                    case Success(value):
                        return Success.from_value(value)
            elif isinstance(found, Container.ResolvableList):
                match found.retrieve(field, **kwargs):
                    case Success(value):
                        return Success.from_value(value)
            else:
                return Success.from_value(found)

        return Failure(None)

        # elif inspect.isclass(field):

        #     match self._resolve_arguments(field, **kwargs):
        #         case Success(value):
        #             if value == {}:
        #                 return Success.from_value(field())
        #             return Success.from_value(field(**value))
        #     for k, v in self._resolvable.items():
        #         if isinstance(v, Container.ResolvableList):
        #             if v.contains(field):
        #                 return v.retrieve(field)

        #     return Failure(f"Unable to resolve arguments to {field}.__init__")
        # elif Container.is_optional_type(field):
        #     return self._resolve_optional_argument(field)
        # else:
        #     return self._resolve_argument(field)

    def _resolve_arguments(self, fn: callable, **kwargs) -> Result[dict, None]:
        if inspect.isclass(fn):
            required_args = get_type_hints(fn.__init__)
        else:
            required_args = get_type_hints(fn)
        resolved_kwargs = {}

        _log.debug(f"Required args are: {required_args}")
        if len(required_args) == 0:
            return Success({})

        for k, v in required_args.items():
            try:
                resolved_kwargs[k] = kwargs.pop(k)
            except KeyError:
                match self._resolve_argument(v):
                    case Success(value):
                        resolved_kwargs[k] = value
                    case Failure(_):
                        return Failure(f"Missing {k} argument")

        return Success(resolved_kwargs)

    class ResolvableList:

        def __init__(self, type: T):
            self._the_type = type
            self._resolvers: dict[S, Container.Singleton] = {}

        def find_subclass(self, type: T) -> Optional[S]:
            """Find subclass of type T in the resolvers.

            :param type: Type to search for
            :type type: T
            :return: The subclass if found or None
            :rtype: Optional[S]
            """
            for k, v in self._resolvers.items():
                if issubclass(k, type):
                    return k
            return None

        @property
        def type(self) -> T:
            return self._the_type

        def retrieve(self, type: S) -> Result[S, None]:
            if type not in self._resolvers:
                return Failure(None)
            return self._resolvers[type].resolve()

        def contains(self, type: S) -> bool:
            return type in self._resolvers

        def add_singleton(self, type: S, singleton: Container.Singleton):
            if type in self._resolvers:
                raise ValueError(
                    f"Can only have one resolver of {S} in ResolvableList of type {self._the_type}"
                )
            self._resolvers[type] = singleton

    class SingletonMulti:

        def __init__(self, value: Container.Singleton, **kwargs):
            _log.debug(f"Creating singlton value: {S} with kwargs: {kwargs}")
            self._value = value
            self._resolved = None
            self._kwargs = kwargs

        @property
        def value(self):
            if not self._resolved:
                self._resolved = self._value.ValueError
            return self._resolved

    class Singleton(Resolvable):

        def __init__(self, container: Container, value: S, **kwargs):
            _log.debug(f"Creating singlton value: {S} with kwargs: {kwargs}")
            self._value = value
            self._resolved = None
            self._kwargs = kwargs
            self._container: Container = container

            # self._kwargs = {k: v for k, v in kwargs.items()}

        @property
        def type(self) -> S:
            if not inspect.isclass(self._value):
                return type(self._value)
            return type(self._value)

        @property
        def instance(self) -> S:
            if self._resolved is not None:
                match self._container._resolve_arguments(self._value, **self._kwargs):
                    case Success(value):
                        self._resolved = value
                    case Failure(_):
                        raise Unresolvable(self._value)

        @property
        def value(self) -> Maybe[S]:
            if not self._resolved:
                _log.debug(f"Resolving: {self._value}")
                self._resolved = self.resolved(**self._kwargs)
            return self._resolved

        def resolve(self) -> Result[S, None]:
            if not self._resolved:
                self._resolved = self._do_resolution()

            if not self._resolved:
                return Failure(None)

            return Success(self._resolved)

        def _do_resolution(self, **kwargs) -> Result[S, None]:
            if not self._resolved:
                self._kwargs.update(kwargs)
                if 'kwargs' in self._kwargs and self._kwargs['kwargs'] == {}:
                    del self._kwargs['kwargs']

                # Look for args that can be satisfied by looking up
                # the resovable services.
                resolved_kwargs: Result[dict, None] = self._container._resolve_arguments(
                    self._value.__init__, **self._kwargs)

                # required_args = get_type_hints(self._value.__init__)

                # _log.debug(f"Required args are: {required_args}")
                # for k, v in required_args.items():
                #     non_none_type = v

                #     # Handle optional
                #     if typing.get_origin(v) == typing.Union and len(typing.get_args(v)) > 2:
                #         raise ResolutionError(f"Argument {v} to resolve {k} has more than 2 types")
                #     elif typing.get_origin(v) == typing.Union and None not in typing.get_args(v):
                #         raise ResolutionError(
                #             f"Argument {v} to resolve {k} only None and one more type is allowed to resolve")
                #     elif typing.get_origin(v) == typing.Union:
                #         non_none_type = next(filter(lambda l: l is not None, typing.get_args(v)))

                #     if non_none_type in self._resolvable:
                #         if isinstance(self._resolvable[non_none_type], Container.Singleton) or isinstance(
                #                 self._resolvable[non_none_type], Container.Transient):
                #             _log.debug(
                #                 f"For k: {k} using: {self._resolvable[non_none_type]._value} to resolve (.value) for.")
                #             self._kwargs[k] = self._resolvable[non_none_type].value
                #         else:
                #             _log.debug(f"For k: {k} using: {self._resolvable[non_none_type]} to resolve (.value) for.")
                #             self._kwargs[k] = self._resolvable[non_none_type]

                try:
                    arg_dict = resolved_kwargs.unwrap()
                    self._resolved = self._value(**resolved_kwargs.unwrap())
                except TypeError as ex:
                    for x in ex.args:
                        if 'kwargs' in x or 'no arguments' in x:
                            self._resolved = self._value()
                            break
                    if not self._resolved:
                        raise ex

            return self._resolved

    class Transient:

        def __init__(self, value: S, **kwargs):
            self._value = value
            self._kwargs = kwargs

        @property
        def value(self):
            return self._resolve_kwargs(**self._kwargs)
            # return self._value(**self._kwargs)

        def _resolve_kwargs(self, **kwargs) -> S:
            # Look for args that can be satisfied by looking up
            # the resovable services.
            required_args = get_type_hints(self._value.__init__)

            my_kwargs = kwargs
            ret_value: S = None

            for k, v in required_args.items():
                to_resolve = v
                if Container.is_optional_type(v):
                    if len(typing.get_args(v)) > 2:
                        raise ValueError("Only support optional of single values")
                    for option in typing.get_args(v):
                        if option is not None:
                            to_resolve = option
                            break
                    _log.debug(f"k: {k} is optional {typing.get_origin(v)}")
                if to_resolve in self._resolvable:
                    if isinstance(self._resolvable[to_resolve], Container.Singleton) or isinstance(
                            self._resolvable[to_resolve], Container.Transient):
                        _log.debug(
                            f"For k: {k} using: {self._resolvable[to_resolve]._value} to resolve (.value) for."
                        )
                        self._kwargs[k] = self._resolvable[to_resolve].value
                    else:
                        _log.debug(
                            f"For k: {k} using: {self._resolvable[to_resolve]} to resolve (.value) for."
                        )
                        self._kwargs[k] = self._resolvable[to_resolve]

            try:
                ret_value = self._value(**self._kwargs)
            except TypeError as ex:
                for x in ex.args:
                    if 'kwargs' in x:
                        ret_value = self._value()
                        break
                if not ret_value:
                    raise ex
            return ret_value

    # def add_singleton_with_many(self, type: T, value: S, **kwargs: dict):
    # def add_singleton_by_name()

    def add_interface_reference(self, type: T, value: S, **kwargs: dict):
        """
        Add a reference where type T resolves to value S.  In this case the type can
        be a Protocol, BaseClass or any class which value: S implements.  And will be
        resolvable only by resolve(T).

        :param type: An interface for using to resolve to
        :type type: T
        :param value: An object instance or resolvable class for interface retrieval.
        :type value: S
        :raises AttributeError: Raised if a type T already has been added to the resolved items.
        """
        if type in self._resolvable:
            raise AttributeError(f"Type: {type} already exists and can't be modified.")
        self._resolvable[type] = Container.Singleton(self, value, **kwargs)

    def add_concrete_reference(self, type: T, value: S, **kwargs: dict):

        resolvable_list: Container.ResolvableList = None
        if type in self._resolvable:
            if not isinstance(self._resolvable[type], Container.ResolvableList):
                raise AttributeError(f"{type} already found as an instance_reference.")
            resolvable_list: Container.ResolvableList = self._resolvable[type]
            if resolvable_list.contains(value):
                raise ValueError(f"{value} already within the resolvable list type {T} ")
        else:
            resolvable_list = Container.ResolvableList(type=type)
            self._resolvable[type] = resolvable_list
        if not type in self._resolvable_lists:
            self._resolvable_lists[type] = resolvable_list
        resolvable_list.add_singleton(value, Container.Singleton(self, value, **kwargs))

    def add_instance(self, type: T, value: S):
        self._resolvable[type] = value

    def add_factory(self, type: T, value: S, **kwargs):
        self._resolvable[type] = Container.Transient(value, **kwargs)

    def resolve(self, type: T, **kwargs) -> Optional[S]:
        _log.debug(f"Attempting resolve of type: {type}")

        match self._resolve(type, **kwargs):
            case Success(value):
                self._resolvable[type] = value
                return value
            case Failure(_):
                raise Unresolvable(type)

        return None

        resolved = pipe(self._resolve)
        return resolved(type)

        try:
            unwrapped = resolved.unwrap()
            if unwrapped is None:
                raise Unresolvable(type=type)
            else:
                return unwrapped
        except returns.primitives.exceptions.UnwrapFailedError:
            raise Unresolvable(type=type)

        return resolved.unwrap_or(None)

        try:
            resolved: Maybe = None
            obj = self.resolvable[type]
            if isinstance(obj, Container.Transient):
                resolved = obj.resolve()
                return obj.value()
            elif isinstance(obj, Container.Singleton):
                return obj.resolved()
            return self.resolvable[type]
        except KeyError:
            for k, v in self.resolvable.items():
                if isinstance(v, self._resolvableList):
                    if v.contains(type):
                        return v.retrieve(type)
            raise Unresolvable(type=type)

    @staticmethod
    def create(name: str) -> Container:
        if name in Container.containers:
            raise ValueError(f"A container named: {name} already exists")
        container = Container(name=name)
        Container.containers[name] = container
        return container

    @staticmethod
    def get_container(name: str) -> Container:
        return Container.containers[name]


service_repo = Container()



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2023 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import argparse
import os
import sys
import tempfile
import subprocess
from volttron.utils import jsonapi

from volttron.client.known_identities import CONFIGURATION_STORE
from volttron.utils.jsonrpc import RemoteError

_stdout = sys.stdout
_stderr = sys.stderr


def add_config_to_store(opts):
    opts.connection.peer = CONFIGURATION_STORE
    call = opts.connection.call

    file_contents = opts.infile.read()

    call(
        "set_config",
        opts.identity,
        opts.name,
        file_contents,
        config_type=opts.config_type,
    )


def delete_config_from_store(opts):
    opts.connection.peer = CONFIGURATION_STORE
    call = opts.connection.call
    if opts.delete_store:
        call("delete_store", opts.identity)
        return

    if opts.name is None:
        _stderr.write("ERROR: must specify a configuration when not deleting entire "
                      "store\n")
        return

    call("delete_config", opts.identity, opts.name)


def list_store(opts):
    opts.connection.peer = CONFIGURATION_STORE
    call = opts.connection.call
    results = []
    if opts.identity is None:
        results = call("list_stores")
    else:
        results = call("list_configs", opts.identity)

    for item in results:
        _stdout.write(item + "\n")


def get_config(opts):
    opts.connection.peer = CONFIGURATION_STORE
    call = opts.connection.call
    results = call("get_config", opts.identity, opts.name, raw=opts.raw)

    if opts.raw:
        _stdout.write(results)
    else:
        if isinstance(results, str):
            _stdout.write(results)
        else:
            _stdout.write(jsonapi.dumps(results, indent=2))
            _stdout.write("\n")


def edit_config(opts):
    opts.connection.peer = CONFIGURATION_STORE
    call = opts.connection.call

    if opts.new_config:
        config_type = opts.config_type
        raw_data = ""
    else:
        try:
            results = call("get_metadata", opts.identity, opts.name)
            config_type = results["type"]
            raw_data = results["data"]
        except RemoteError as e:
            if "No configuration file" not in e.message:
                raise
            config_type = opts.config_type
            raw_data = ""

    # Write raw data to temp file
    # This will not work on Windows, FYI
    with tempfile.NamedTemporaryFile(suffix=".txt", mode="r+") as f:
        f.write(raw_data)
        f.flush()

        success = True
        try:
            # do not use utils.execute_command as we don't want set stdout to
            #  subprocess.PIPE
            subprocess.check_call([opts.editor, f.name])
        except subprocess.CalledProcessError as e:
            _stderr.write("Editor returned with code {}. Changes not "
                          "committed.\n".format(e.returncode))
            success = False

        if not success:
            return

        f.seek(0)
        new_raw_data = f.read()

        if new_raw_data == raw_data:
            _stderr.write("No changes detected.\n")
            return

        call(
            "set_config",
            opts.identity,
            opts.name,
            new_raw_data,
            config_type=config_type,
        )


def add_config_store_parser(add_parser_fn):
    config_store = add_parser_fn("config", help="manage the platform configuration store")
    config_store_subparsers = config_store.add_subparsers(title="subcommands",
                                                          metavar="",
                                                          dest="store_commands")

    config_store_store = add_parser_fn("store",
                                       help="store a configuration",
                                       subparser=config_store_subparsers)

    config_store_store.add_argument("identity", help="VIP IDENTITY of the store")
    config_store_store.add_argument(
        "name", help="name used to reference the configuration by in the store")
    config_store_store.add_argument(
        "infile",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="file containing the contents of the configuration",
    )
    config_store_store.add_argument(
        "--raw",
        const="raw",
        dest="config_type",
        action="store_const",
        help="interpret the input file as raw data",
    )
    config_store_store.add_argument(
        "--json",
        const="json",
        dest="config_type",
        action="store_const",
        help="interpret the input file as json",
    )
    config_store_store.add_argument(
        "--csv",
        const="csv",
        dest="config_type",
        action="store_const",
        help="interpret the input file as csv",
    )

    config_store_store.set_defaults(func=add_config_to_store, config_type="json")

    config_store_edit = add_parser_fn(
        "edit",
        help="edit a configuration. (nano by default, respects EDITOR env "
        "variable)",
        subparser=config_store_subparsers,
    )

    config_store_edit.add_argument("identity", help="VIP IDENTITY of the store")
    config_store_edit.add_argument("name",
                                   help="name used to reference the configuration by in the store")
    config_store_edit.add_argument(
        "--editor",
        dest="editor",
        help="Set the editor to use to change the file. Defaults to nano if "
        "EDITOR is not set",
        default=os.getenv("EDITOR", "nano"),
    )
    config_store_edit.add_argument(
        "--raw",
        const="raw",
        dest="config_type",
        action="store_const",
        help="Interpret the configuration as raw data. If the file already "
        "exists this is ignored.",
    )
    config_store_edit.add_argument(
        "--json",
        const="json",
        dest="config_type",
        action="store_const",
        help="Interpret the configuration as json. If the file already "
        "exists this is ignored.",
    )
    config_store_edit.add_argument(
        "--csv",
        const="csv",
        dest="config_type",
        action="store_const",
        help="Interpret the configuration as csv. If the file already exists "
        "this is ignored.",
    )
    config_store_edit.add_argument(
        "--new",
        dest="new_config",
        action="store_true",
        help="Ignore any existing configuration and creates new empty file."
        " Configuration is not written if left empty. Type defaults to "
        "JSON.",
    )

    config_store_edit.set_defaults(func=edit_config, config_type="json")

    config_store_delete = add_parser_fn("delete",
                                        help="delete a configuration",
                                        subparser=config_store_subparsers)
    config_store_delete.add_argument("identity", help="VIP IDENTITY of the store")
    config_store_delete.add_argument(
        "name",
        nargs="?",
        help="name used to reference the configuration by in the store",
    )
    config_store_delete.add_argument(
        "--all",
        dest="delete_store",
        action="store_true",
        help="delete all configurations in the store",
    )

    config_store_delete.set_defaults(func=delete_config_from_store)

    config_store_list = add_parser_fn(
        "list",
        help="list stores or configurations in a store",
        subparser=config_store_subparsers,
    )

    config_store_list.add_argument("identity", nargs="?", help="VIP IDENTITY of the store to list")

    config_store_list.set_defaults(func=list_store)

    config_store_get = add_parser_fn(
        "get",
        help="get the contents of a configuration",
        subparser=config_store_subparsers,
    )

    config_store_get.add_argument("identity", help="VIP IDENTITY of the store")
    config_store_get.add_argument("name",
                                  help="name used to reference the configuration by in the store")
    config_store_get.add_argument("--raw",
                                  action="store_true",
                                  help="get the configuration as raw data")
    config_store_get.set_defaults(func=get_config)



from __future__ import annotations    # Allows reference to current class

from abc import ABC, abstractmethod, abstractstaticmethod
from dataclasses import dataclass
from typing import Protocol, runtime_checkable
import os
from pathlib import Path

from dataclass_wizard import JSONSerializable

from volttron.utils import jsonapi


class CredentialStoreError(Exception):
    pass


class IdentityNotFound(Exception):
    pass


class IdentityAlreadyExists(Exception):
    pass


class InvalidCredentials(Exception):
    pass


@dataclass(frozen=True, kw_only=True)
class Credentials(JSONSerializable):
    identity: str

    def create(*, identity: str) -> Credentials:
        return Credentials(identity=identity)


@dataclass(frozen=True, kw_only=True)
class PublicCredentials(Credentials):
    publickey: str

    def create(*, identity: str, publickey: str) -> PublicCredentials:
        return PublicCredentials(identity=identity, publickey=publickey)


@dataclass(frozen=True, kw_only=True)
class PKICredentials(PublicCredentials):
    secretkey: str

    @property
    def type(self):
        return self.__class__

    def create(*, identity: str, publickey: str, secretkey: str) -> PKICredentials:
        return PKICredentials(identity=identity, publickey=publickey, secretkey=secretkey)

    def create_with_generator(*, identity: str, generator_fn: callable) -> PKICredentials:
        publickey, secretkey = generator_fn()
        return PKICredentials(identity=identity, publickey=publickey, secretkey=secretkey)


@dataclass(frozen=True, kw_only=True)
class VolttronCredentials(PKICredentials):
    domain: str = 'VIP'
    address: str = ''

    @property
    def type(self):
        return self.__class__

    @staticmethod
    def load_from_file(filename: str | Path) -> Credentials:

        if filename is None:
            raise ValueError(f"filename cannot be None")

        if isinstance(filename, str):
            filename = Path(filename).expanduser()

        filename = filename.absolute()

        if not filename.exists():
            raise ValueError(f"filename: {filename} does not exist.")

        obj = jsonapi.loads(filename.read_text())

        return VolttronCredentials.from_dict(obj)


# @service
class CredentialsFactory:
    # def __init__(self, server_options: ServerOptions):
    #     server_options.
    #     CredentialsFactory.CREDENTIAL_STORE = os.environ

    @staticmethod
    def load_from_environ() -> Credentials:
        if credentials := os.environ.get("AGENT_CREDENTIALS"):
            # Expand user variables
            credentials = os.path.expanduser(credentials)
            try:
                creds = CredentialsFactory.load_credentials_from_file(credentials)
                return creds
            except FileNotFoundError:

                # Attempt to load from the environmental variable.
                obj = jsonapi.loads(credentials)

                creds = VolttronCredentials.from_dict(obj)
                return creds

        raise ValueError("No AGENT_CREDENTIALS Environmental Variable")

    @staticmethod
    def load_credentials_from_file(path: Path | str) -> Credentials:
        """
        Create a `Credentials` object from the specified path.

        This function reads from a file and attempts to parse and load the Credentials
        object from that file.  We only support json based credential files.  If the
        credential file holds a public and secret attribute key then a PublicKeyCredential is
        loaded, otherwise a basic Credentials object is loaded.

        :param identity: The identity that should be passed into the Credentals
        :type identity: str
        :param path: A path for the Credentials to load from
        :type path: Path | str
        :raises FileNotFoundError: If path does not exist.
        :return: A credentials object or raises an exception.
        :rtype: Credentials
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise FileNotFoundError(f"Credential file: {path} not found!")

        obj = jsonapi.load(path.open())

        # TODO: Handle this better using type within the keystore.json file.
        if "publickey" in obj and "secretkey" in obj:
            return PKICredentials.create(identity=obj["identity"],
                                         publickey=obj["publickey"],
                                         secretkey=obj["secretkey"])
        else:
            return Credentials.create(identity=obj["identity"])


class CredentialsCreator(ABC):

    @abstractmethod
    def create(*, identity: str, **kwargs) -> Credentials:
        ...

    # def delete(*, identity: str) -> None:
    #     ...

    # def getall() -> list[Credentials]:
    #     ...


class DefaultCredentialsFactory(CredentialsCreator):

    def create(*, identity: str) -> Credentials:
        return Credentials(identity=identity)


class DefaultPKICredentialsFactory(CredentialsCreator):

    def create(*, identity: str, publickey: str, secretkey: str) -> Credentials:
        return PKICredentials(identity=identity, publickey=publickey, secretkey=secretkey)


class CredentialsStore(ABC):

    @abstractmethod
    def get_credentials_type(self) -> type:
        ...

    @abstractmethod
    def store_credentials(self, *, credentials: Credentials) -> None:
        """
        Store credentials for an identity.

        :param identity: The identity to store credentials for.
        :type identity: str
        :param credentials: The credentials to store.
        :type credentials: Credentials
        :raises: IdentityAlreadyExists: If the identity alredy exists, an IdentityAlreadyExists exception MUST be raised.
        """
        ...

    @abstractmethod
    def retrieve_credentials(self, **kwargs) -> Credentials | None:
        """
        Retrieve credentials based upon passed criteria.

        It is up to the implementor to make sure that the passed kwargs are
        processed correctly and return the correct response.
        """
        ...

    @abstractmethod
    def remove_credentials(self, *, identity: str) -> None:
        """
        Delete the credentials for an identity.

        :param identity: The identity to delete credentials for.
        :type identity: str
        :raises: IdentityNotFound: If the identity does not exist, an IdentityNotFound exception MUST be raised.
        """
        ...



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from string import whitespace

store_ext = ".store"
link_prefix = "config://"


def strip_config_name(config_name):
    return config_name.strip(whitespace + r"\/")


def check_for_config_link(value):
    if value.startswith(link_prefix):
        config_name = value.replace(link_prefix, "", 1)
        config_name = strip_config_name(config_name)
        return config_name.lower()
    return None


def list_unique_links(config):
    """Returns a set of config files referenced in this configuration"""
    results = set()
    if isinstance(config, dict):
        values = list(config.values())
    elif isinstance(config, list):
        values = config
    else:
        # Raw config has no links
        return results

    for value in values:
        if isinstance(value, (dict, list)):
            results.update(list_unique_links(value))
        elif isinstance(value, str):
            value = value.lower()
            if value.startswith(link_prefix):
                config_name = value.replace(link_prefix, "", 1)
                config_name = strip_config_name(config_name)
                results.add(config_name)

    return results


def check_for_recursion(new_config_name, new_config, existing_configs):
    return _follow_links(
        set(),
        new_config_name.lower(),
        new_config_name.lower(),
        new_config,
        existing_configs,
    )


def _follow_links(
    seen,
    new_config_name,
    current_config_name,
    current_config,
    existing_configs,
):
    children = list_unique_links(current_config)

    if new_config_name in children:
        return True

    seen.add(current_config_name)

    for child_config_name in children - seen:
        child_config = existing_configs.get(child_config_name)
        if child_config is None:
            # Link to a non-existing config, skip in the future.
            seen.add(child_config_name)
            continue
        if _follow_links(
                seen,
                new_config_name,
                child_config_name,
                child_config,
                existing_configs,
        ):
            return True

    return False



from __future__ import annotations

import os
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from dataclass_wizard import JSONSerializable, JSONWizard

from volttron.types.blinker_events import volttron_home_set_evnt

PropertyValue = str | int | float


@dataclass
class _KnownHostProperties(JSONWizard):
    hosts: dict[str, dict] = field(default_factory=dict)

    def add_property(self, host: str, key: str, value: PropertyValue):
        if host not in self.hosts:
            self.hosts[host] = dict()
        self.hosts[host][key] = value

    def has_property(self, host: str, key: str) -> bool:
        return key in self.hosts[host]

    def get_property(self, host: str, key: str) -> PropertyValue:
        return self.hosts[host][key]

    def store(self, path: Path | str):
        if isinstance(path, str):
            path = Path(path)
        path.open("w").write(self.to_json(indent=2))

    def get_host_properties(self, host: Optional[str]) -> dict[str, PropertyValue]:
        if host is None:
            return self.hosts["@"]
        return self.hosts.get(host, dict())

    @staticmethod
    def load(path: Path) -> KnownHostProperties:
        if isinstance(path, str):
            path = Path(path)

        if path.exists():
            obj = _KnownHostProperties.from_json(path.open().read())
        else:
            obj = _KnownHostProperties()
        return obj


KnownHostProperties = None


@volttron_home_set_evnt.connect
def volttron_home(sender: Any):
    global KnownHostProperties
    volttron_home = os.environ.get("VOLTTRON_HOME")
    assert volttron_home, "Requires VOLTTRON_HOME to be set"
    KnownHostProperties = _KnownHostProperties.load(Path(volttron_home) / "known_hosts.json")


if __name__ == '__main__':
    import json

    old = json.loads("""{
    "@": "2YYmfjduH-D2SzPZO85dkubqv3aCJlJTfB_C514LXRA",
    "127.0.0.1:22916": "2YYmfjduH-D2SzPZO85dkubqv3aCJlJTfB_C514LXRA"
    }""")

    p = KnownHostProperties()
    p.add_property("@", "publickey", old["@"])
    p.add_property("@", "publickey", old["@"])
    p.add_property("127.0.0.1:22916", "publickey", old["127.0.0.1:22916"])

    print(p.to_json(indent=2))
    p.store("foo.json")

    b = KnownHostProperties.load("foo.json")

    print(b.to_json(indent=2))



import copy
import logging
import re
from typing import Any, TYPE_CHECKING

from attrs import validators, define, field, fields
from cattrs import Converter

from volttron.types import Identity

vipid_dot_rpc_method = str
RPC_CAPABILITIES = "rpc_capabilities"
PUBSUB_CAPABILITIES = "pubsub_capabilities"
ROLES = "roles"
AGENTS = "agents"
GROUPS = "agent_groups"
IDENTITIES = "identities"
authz_converter = Converter()
_log = logging.getLogger(__name__)


@define
class RPCCapability:
    resource = field(type=vipid_dot_rpc_method)
    param_restrictions = field(type=dict, default=None)

    def __attrs_post_init__(self):
        if self.param_restrictions is None:
            self.param_restrictions = dict()

    def add_param_restrictions(self, param: str, value: Any):
        self.param_restrictions[param] = value


@define
class RPCCapabilities:
    rpc_capabilities = field(type=list[RPCCapability], default=None)
    _rpc_dict = field(type=dict, default=None, init=False)

    def __attrs_post_init__(self):
        if self.rpc_capabilities is None:
            self.rpc_capabilities = []
        self._rpc_dict = dict()
        for r in self.rpc_capabilities:
            self._rpc_dict[r.resource] = r.param_restrictions

    def add_rpc_capability(self, c: RPCCapability):
        if c.resource not in self._rpc_dict:
            self.rpc_capabilities.append(c)
            self._rpc_dict[c.resource] = c.param_restrictions
        else:
            for r in self.rpc_capabilities:
                if r.resource == c.resource:
                    r.param_restrictions.update(c.param_restrictions)
                    self._rpc_dict[r.resource] = r.param_restrictions
                    break

    def remove_rpc_capability(self, c: RPCCapability):
        try:
            self.rpc_capabilities.remove(c)
        except ValueError:
            if c.resource in self._rpc_dict:
                # difference is in param_restriction
                for r in self.rpc_capabilities:
                    if r.resource == c.resource:
                        for k in c.param_restrictions:
                            r.param_restrictions.pop(k, None)
        else:
            self._rpc_dict.pop(c.resource)

    def __len__(self):
        return len(self.rpc_capabilities)


def unstructure_rpc_capabilities(instance: RPCCapabilities):
    """
    Convert from:

        RPCCapabilities(
        rpc_capabilities=[RPCCapability(resource='id.rpc1', param_restrictions={}),
                          RPCCapability(resource='id2.rpc2', param_restrictions={'id': 'id2', 'p2': 'v2'})])

    TO:
        ['id.rpc1', {'id2.rpc2': {'id': 'id2', 'p2': 'v2'}}]
        i.e. instead of the default unstructure/asdict behavior - list of {resource:value, param_restrictions:value}
        generate single List with just resource str if param_restrictions is None or dict(resource, param_restriction)

    """
    rpc_capabilities_list = []
    for c in instance.rpc_capabilities:
        if c.param_restrictions:
            rpc_capabilities_list.append({c.resource: c.param_restrictions})
        else:
            rpc_capabilities_list.append(c.resource)
    return rpc_capabilities_list


@define
class PubsubCapability:
    topic_pattern = field(type=str)
    topic_access = field(validator=validators.in_(["publish", "subscribe", "pubsub"]))


@define
class PubsubCapabilities:
    pubsub_capabilities = field(type=list[PubsubCapability], default=None)

    def __attrs_post_init__(self):
        if self.pubsub_capabilities is None:
            self.pubsub_capabilities = []

    def add_pubsub_capability(self, c: PubsubCapability):
        if c not in self.pubsub_capabilities:
            self.pubsub_capabilities.append(c)

    def remove_pubsub_capability(self, c: PubsubCapability):
        try:
            self.pubsub_capabilities.remove(c)
        except ValueError:
            pass

    def __len__(self):
        return len(self.pubsub_capabilities)


def unstructure_pubsub_capabilities(instance: PubsubCapabilities):
    """
    Convert from:
        PubsubCapabilities(
        pubsub_capabilities=[PubsubCapability(topic_pattern='device/*', topic_access='pubsub'),
                             PubsubCapability(topic_pattern='*', topic_access='pubsub')])
    TO:
        {'device/*': 'pubsub', '*': 'pubsub'}
        i.e. instead of the default - list of dict with topic_pattern:<value>, topic_access: <value>
             generate single dict with key as topic_pattern and values topic_access
    """
    pubsub_capabilities_dict = dict()
    for c in instance.pubsub_capabilities:
        pubsub_capabilities_dict[c.topic_pattern] = c.topic_access
    return pubsub_capabilities_dict


@define
class Role:
    name = field(type=str)
    rpc_capabilities = field(type=RPCCapabilities, default=None)
    pubsub_capabilities = field(type=PubsubCapabilities, default=None)

    def add_rpc_capability(self, c: RPCCapability):
        self.rpc_capabilities.add_rpc_capability(c)

    def remove_rpc_capability(self, c: RPCCapability):
        self.rpc_capabilities.remove_rpc_capability(c)


@define
class Roles:
    roles = field(type=list[Role], default=None)

    def __attrs_post_init__(self):
        if self.roles is None:
            self.roles = []

    def __len__(self):
        return len(self.roles)


def unstructure_roles(instance: Roles):
    """
    Convert from:
        Roles(roles=[
          Role(name='admin',
               rpc_capabilities=RPCCapabilities(rpc_capabilities=[
                                 RPCCapability(resource='id.rpc1', param_restrictions={}),
                                 RPCCapability(resource='id2.rpc2', param_restrictions={'id': 'id2', 'p2': 'v2'})]),
               pubsub_capabilities=None),
          Role(name='new_role',
               rpc_capabilities=None,
               pubsub_capabilities=PubsubCapabilities(pubsub_capabilities=[
                                     PubsubCapability(topic_pattern='device/*', topic_access='pubsub'),
                                     PubsubCapability(topic_pattern='*', topic_access='pubsub')]))
          ])

    TO:
        {'admin': {'rpc_capabilities': ['id.rpc1', {'id2.rpc2': {'id': 'id2', 'p2': 'v2'}}]},
         'new_role': {'pubsub_capabilities': {'device/*': 'pubsub', '*': 'pubsub'}}}

        i.e. instead of list of dict with name:<value>, rpc_capabilities: <list>, pubsub_capabilities: <list>
             generate dict with key as role name and values as rpc and pubsub capabilities and skip none values
    """
    _roles_dict = dict()
    for _r in instance.roles:
        _roles_dict[_r.name] = dict()
        if _r.rpc_capabilities:
            _roles_dict[_r.name][RPC_CAPABILITIES] = authz_converter.unstructure(_r.rpc_capabilities)
        if _r.pubsub_capabilities:
            _roles_dict[_r.name][PUBSUB_CAPABILITIES] = authz_converter.unstructure(_r.pubsub_capabilities)
    return _roles_dict


role_name = str
PubKey = str


@define
class AgentRole:
    role_name = field(type=role_name)
    param_restrictions = field(type=dict, default=None)

    def __attrs_post_init__(self):
        if self.param_restrictions is None:
            self.param_restrictions = dict()

    def add_param_restrictions(self, param: str, value: Any):
        self.param_restrictions[param] = value


@define
class AgentRoles:
    agent_roles = field(type=list[AgentRole], default=None)
    _agent_roles_dict = field(type=dict, default=None, init=False)

    def __attrs_post_init__(self):
        if self.agent_roles is None:
            self.agent_roles = []
        self._agent_roles_dict = dict()
        for r in self.agent_roles:
            self._agent_roles_dict[r.role_name] = r.param_restrictions

    def add_agent_role(self, c: AgentRole):
        if c.role_name not in self._agent_roles_dict:
            self.agent_roles.append(c)
            self._agent_roles_dict[c.role_name] = c.param_restrictions
        else:
            for r in self.agent_roles:
                if r.role_name == c.role_name:
                    r.param_restrictions.update(c.param_restrictions)
                    self._agent_roles_dict[r.role_name] = r.param_restrictions
                    break

    def remove_agent_role(self, c: AgentRole):
        try:
            self.agent_roles.remove(c)
        except ValueError:
            if c.role_name in self._agent_roles_dict:
                # difference is in param_restriction
                for r in self.agent_roles:
                    if r.role_name == c.role_name:
                        for k in c.param_restrictions:
                            r.param_restrictions.pop(k, None)
        else:
            self._agent_roles_dict.pop(c.role_name)

    def __len__(self):
        return len(self.agent_roles)


def unstructure_agent_roles(instance: AgentRoles):
    """
    Convert from:

        AgentRoles(
        agent_roles=[AgentRole(role_name='role1', param_restrictions={}),
                    AgentRole(role_name='role2', param_restrictions={'id': 'id2', 'p2': 'v2'})])

    TO:
        ['role1', {'role2': {'id': 'id2', 'p2': 'v2'}}]
        i.e. instead of the default unstructure/asdict behavior - list of {role_name:value, param_restrictions:value}
        generate single List with just role_name str if param_restrictions is None or dict(role_name, param_restriction)

    """
    agent_roles_list = []
    for c in instance.agent_roles:
        if c.param_restrictions:
            agent_roles_list.append({c.role_name: c.param_restrictions})
        else:
            agent_roles_list.append(c.role_name)
    return agent_roles_list


@define
class AgentGroup:
    name = field(type=str)
    identities = field(type=list[Identity], default=None)
    agent_roles = field(type=AgentRoles, default=None)
    rpc_capabilities = field(type=RPCCapabilities, default=None)
    pubsub_capabilities = field(type=PubsubCapabilities, default=None)

    def __attrs_post_init__(self):
        if self.identities is None:
            self.identities = list()

        if self.agent_roles is None:
            self.agent_roles = AgentRoles()


@define
class AgentGroups:
    agent_groups = field(type=list[AgentGroup], default=None)

    def __attrs_post_init__(self):
        if self.agent_groups is None:
            self.agent_groups = []

    def __len__(self):
        return len(self.agent_groups)


def unstructure_agent_groups(instance: AgentGroups):
    """
    Convert
      AgentGroups(agent_groups=[
        AgentGroup(name='admin_agents',
                  identities=['volttron.ctl', 'config.store'],
                  agent_roles=['admin'],
                  rpc_capabilities=None,
                  pubsub_capabilities=PubsubCapabilities(pubsub_capabilities=[
                    PubsubCapability(topic_pattern='device/*', topic_access='pubsub'),
                    PubsubCapability(topic_pattern='*', topic_access='pubsub')])
                  )
        ])

    TO:
    {'admin_agents':
        {'agents': ['volttron.ctl', 'config.store'],
         'agent_roles': ['admin'],
         'pubsub_capabilities': {'device/*': 'pubsub', '*': 'pubsub'}
         }
    }
    i.e. instead of list of dict with
                  name:<value>, identities:<list>, agent_roles:<list>, rpc_capabilities: <list>, pubsub_capabilities: <list>
    generate dict with key as group name and values as identities, agent_roles, __rpc__, pubsub capabilities and skipping none
    """
    _groups_dict = dict()
    for _g in instance.agent_groups:
        _groups_dict[_g.name] = {'identities': _g.identities}
        if _g.agent_roles:
            _groups_dict[_g.name]['agent_roles'] = authz_converter.unstructure(_g.agent_roles)
        if _g.rpc_capabilities:
            _groups_dict[_g.name][RPC_CAPABILITIES] = authz_converter.unstructure(_g.rpc_capabilities)
        if _g.pubsub_capabilities:
            _groups_dict[_g.name][PUBSUB_CAPABILITIES] = authz_converter.unstructure(_g.pubsub_capabilities)

    return _groups_dict


@define
class Agent:
    identity = field(type=Identity)
    protected_rpcs = field(type=set[vipid_dot_rpc_method], default=None)
    agent_roles = field(type=AgentRoles, default=None)
    rpc_capabilities = field(type=RPCCapabilities, default=None)
    pubsub_capabilities = field(type=PubsubCapabilities, default=None)
    comments = field(type=str, default=None)


@define
class Agents:
    agents = field(type=list[Agent], default=None)

    def __attrs_post_init__(self):
        if self.agents is None:
            self.agents = []

    def __len__(self):
        return len(self.agents)


def unstructure_agents(instance: Agents):
    _agents_dict = dict()
    # make a dict with key as identity and value as dict of all other attributes. In the inner dict, values are
    # converted from attrs objects to primitives based on its own custom unstructure methods
    for _agent in instance.agents:
        _agents_dict[_agent.identity] = dict()
        for attribute in fields(_agent.__class__):
            if attribute.name != "identity" and _agent.__getattribute__(attribute.name):
                _agents_dict[_agent.identity][attribute.name] = (authz_converter.unstructure(
                    _agent.__getattribute__(attribute.name)))
    return _agents_dict


@define
class VolttronAuthzMap:
    protected_topics = field(type=list[str], default=None)
    roles = field(type=Roles, default=None)
    agent_groups = field(type=AgentGroups, default=None)
    agents = field(type=Agents, default=None)
    compact_dict = field(type=dict, init=False, default=None)
    agent_capabilities = field(type=dict, init=False, default=None)

    def __attrs_post_init__(self):
        self.compact_dict = authz_converter.unstructure(self)
        self.agent_capabilities = copy.deepcopy(self.compact_dict.get(AGENTS))
        VolttronAuthzMap.expand_agent_capabilities(agent_capabilities=self.agent_capabilities,
                                                   agent_groups=self.compact_dict.get(GROUPS),
                                                   roles=self.compact_dict.get(ROLES))

    @classmethod
    def from_unstructured_dict(cls, input_dict: dict):
        """
        Method to create VolttronAuthzMap instance from json.load(authz.json file)
        :param input_dict: json.load(authz.json file created by unstructure(VolttronAuthzMap))
        :return: instance of VolttronAuthzMap
        """
        protected_topics = input_dict.get("protected_topics")

        # Build Roles
        _roles = list()
        for name, value in input_dict.get(ROLES, dict()).items():
            rpc_obj_list = VolttronAuthzMap.create_rpc_capabilities_obj(value.get(RPC_CAPABILITIES))
            pubsub_obj_list = VolttronAuthzMap.create_pubsub_capabilities_obj(value.get(PUBSUB_CAPABILITIES))

            r_obj = Role(name, rpc_capabilities=rpc_obj_list, pubsub_capabilities=pubsub_obj_list)
            _roles.append(r_obj)

        # Build agent groups
        groups = list()
        for group_name, value in input_dict.get(GROUPS, dict()).items():
            vip_ids = value.get(IDENTITIES, list())
            role_names = value.get(ROLES, list())
            rpc_obj_list = VolttronAuthzMap.create_rpc_capabilities_obj(value.get(RPC_CAPABILITIES))
            pubsub_obj_list = VolttronAuthzMap.create_pubsub_capabilities_obj(value.get(PUBSUB_CAPABILITIES))
            groups.append(
                AgentGroup(name=group_name,
                           identities=vip_ids,
                           agent_roles=role_names,
                           rpc_capabilities=rpc_obj_list,
                           pubsub_capabilities=pubsub_obj_list))

        # Build agents
        agents = list()
        for identity, value in input_dict.get(AGENTS, dict()).items():
            protected_rpcs = value.get("protected_rpcs", list())
            role_names = value.get('agent_roles', list())
            rpc_obj_list = VolttronAuthzMap.create_rpc_capabilities_obj(value.get(RPC_CAPABILITIES))
            pubsub_obj_list = VolttronAuthzMap.create_pubsub_capabilities_obj(value.get(PUBSUB_CAPABILITIES))
            comments = value.get("comments")
            agents.append(
                Agent(identity=identity,
                      protected_rpcs=protected_rpcs,
                      agent_roles=role_names,
                      rpc_capabilities=rpc_obj_list,
                      pubsub_capabilities=pubsub_obj_list,
                      comments=comments))

        authz_roles = Roles(_roles)
        authz_agent_groups = AgentGroups(groups)
        authz_agents = Agents(agents)
        instance = cls(protected_topics=protected_topics,
                       roles=authz_roles,
                       agent_groups=authz_agent_groups,
                       agents=authz_agents)
        # print(json.dumps(authz.authz_converter.unstructure(instance), indent=4))
        instance.compact_dict = input_dict
        return instance

    @classmethod
    def create_rpc_capabilities_obj(cls, rpc_cap_list: list = None) -> RPCCapabilities:
        if rpc_cap_list is None:
            rpc_cap_list = list()
        obj_list = RPCCapabilities([])    # I don't get a new instance of list in obj if I don't pass []  ?!
        for rpc_cap in rpc_cap_list:
            if isinstance(rpc_cap, str):
                obj_list.add_rpc_capability(RPCCapability(rpc_cap))
            elif isinstance(rpc_cap, dict):
                vip_id_rpc_method = list(rpc_cap.keys())[0]
                param_restrict = rpc_cap[vip_id_rpc_method]
                obj_list.add_rpc_capability(RPCCapability(vip_id_rpc_method, param_restrict))
        return obj_list

    @classmethod
    def create_pubsub_capabilities_obj(cls, pubsub_cap_dict: dict = None) -> PubsubCapabilities:
        if pubsub_cap_dict is None:
            pubsub_cap_dict = dict()
        obj_list = PubsubCapabilities([])    # I don't get a new instance of list in obj if I don't pass []  ?!
        for topic_pattern, access in pubsub_cap_dict.items():
            obj_list.add_pubsub_capability(PubsubCapability(topic_pattern, access))
        return obj_list

    @classmethod
    def expand_agent_capabilities(cls, *, agent_capabilities: dict, agent_groups: dict = None, roles: dict = None):
        if not agent_capabilities:
            return
        apply_role_capabilities = False
        if agent_groups:
            # Apply rules of the group to each group member
            for _name, group_details in agent_groups.items():
                for vip in group_details[IDENTITIES]:
                    cls.update_agent_roles(agent_capabilities[vip], group_details.get("agent_roles"))
                    cls.update_rpc_capabilities(agent_capabilities[vip], group_details.get(RPC_CAPABILITIES))
                    cls.update_pubsub_capabilities(agent_capabilities[vip], group_details.get(PUBSUB_CAPABILITIES))
        if roles or apply_role_capabilities:
            # Apply agent_role's capabilities to agent_capabilities.
            for vip, agent_authz in agent_capabilities.items():
                for agent_role in agent_authz.get("agent_roles", []):
                    if isinstance(agent_role, dict):
                        agent_role_name = list(agent_role.keys())[0]
                        if not roles.get(agent_role_name):
                            # not an agent_role that is currently updated, skip to next agent_role
                            continue
                        param_restriction = agent_role[agent_role_name]
                        # add param restriction dict to role's __rpc__ capbailities before merging with
                        # agent capabilities
                        role_rpc_caps = copy.deepcopy(roles.get(agent_role_name).get(RPC_CAPABILITIES))
                        role_rpc_caps_params = []
                        for role_rpc_cap in role_rpc_caps:
                            if isinstance(role_rpc_cap, str):
                                d = {role_rpc_cap: param_restriction}
                            else:
                                key = list(role_rpc_cap.keys())[0]
                                value = role_rpc_cap[key]
                                d = {key: value}
                                value.update(param_restriction)
                            role_rpc_caps_params.append(d)
                        # Update agent authz with role's __rpc__ cap
                        cls.update_rpc_capabilities(agent_authz, role_rpc_caps_params)
                    else:
                        agent_role_name = agent_role
                        if not roles.get(agent_role_name):
                            # not a role that is currently updated, skip to next role
                            continue
                        # Update agent authz with role's __rpc__ cap
                        cls.update_rpc_capabilities(agent_authz, roles.get(agent_role_name).get(RPC_CAPABILITIES))

                    # update agent authz with role's pubsub cap
                    cls.update_pubsub_capabilities(agent_authz, roles.get(agent_role_name).get(PUBSUB_CAPABILITIES))

    @classmethod
    def update_rpc_capabilities_or_roles(cls, authz_dict: dict, new_caps_list: list, list_type: str):

        if not new_caps_list:
            return
        current_caps_list = authz_dict.get(list_type)
        if not current_caps_list:
            authz_dict[list_type] = copy.deepcopy(new_caps_list)
            return

        # Both current and new __rpc__ capabilities are not empty.
        # Merge
        for new_cap in new_caps_list:
            if isinstance(new_cap, dict):
                new_cap_name = next(iter(new_cap))
            else:
                new_cap_name = new_cap
            for i, cur_cap in enumerate(current_caps_list):
                if isinstance(cur_cap, dict) and new_cap_name == next(iter(cur_cap)):
                    if isinstance(new_cap, dict):
                        new_param_restrict_value = copy.deepcopy(new_cap[new_cap_name])
                        new_param_restrict_value.update(cur_cap[new_cap_name])
                        current_caps_list[i] = {new_cap_name: new_param_restrict_value}
                        break
                    else:
                        current_caps_list[i] = new_cap_name
                        break
                elif cur_cap == new_cap_name:
                    if isinstance(new_cap, dict):
                        current_caps_list[i] = {cur_cap: copy.deepcopy(new_cap[cur_cap])}
                        break
                    else:
                        # both are string and capability already exists do nothing
                        break
            else:
                # for loop completed and didn't do anything so new cap is not in current. so append
                current_caps_list.append(copy.deepcopy(new_cap))

    @classmethod
    def update_rpc_capabilities(cls, authz_dict: dict, new_rpc_caps: list):
        cls.update_rpc_capabilities_or_roles(authz_dict, new_rpc_caps, RPC_CAPABILITIES)

    @classmethod
    def update_agent_roles(cls, authz_dict: dict, new_roles: list):
        cls.update_rpc_capabilities_or_roles(authz_dict, new_roles, 'agent_roles')

    @classmethod
    def update_pubsub_capabilities(cls, authz_dict, new_pubsub_caps):
        if new_pubsub_caps:
            agent_pubsub_caps = authz_dict.get(PUBSUB_CAPABILITIES, dict())
            agent_pubsub_caps.update(new_pubsub_caps)
            authz_dict[PUBSUB_CAPABILITIES] = agent_pubsub_caps

    def create_or_merge_role(self,
                             *,
                             name: str,
                             rpc_capabilities: RPCCapabilities = None,
                             pubsub_capabilities: PubsubCapabilities = None) -> bool:
        if not rpc_capabilities and not pubsub_capabilities:
            raise ValueError(f"Role {name} should have non empty capabilities - rpc capabilities, "
                             "pubsub capabilities or both")
        if not self.compact_dict.get(ROLES):
            self.compact_dict[ROLES] = dict()
        if name not in self.compact_dict.get(ROLES):
            self.compact_dict.get(ROLES)[name] = dict()

        expand_agent_caps = False
        role_dict = self.compact_dict.get(ROLES).get(name)
        if role_dict:
            expand_agent_caps = True    # existing role so might have agents associated it, updated agent_caps
        VolttronAuthzMap.update_rpc_capabilities(role_dict, authz_converter.unstructure(rpc_capabilities))
        VolttronAuthzMap.update_pubsub_capabilities(role_dict, authz_converter.unstructure(pubsub_capabilities))

        if expand_agent_caps:
            self.expand_agent_capabilities(agent_capabilities=self.agent_capabilities, roles={name: role_dict})
        return True

    def create_or_merge_agent_group(self,
                                    *,
                                    name: str,
                                    identities: list[Identity],
                                    agent_roles: AgentRoles = None,
                                    rpc_capabilities: RPCCapabilities = None,
                                    pubsub_capabilities: PubsubCapabilities = None,
                                    **kwargs) -> bool:
        new_group = False
        if not self.compact_dict.get(GROUPS):
            self.compact_dict[GROUPS] = dict()
        if name not in self.compact_dict.get(GROUPS):
            self.compact_dict.get(GROUPS)[name] = dict()
            new_group = True

        if new_group and not identities:
            self.compact_dict.get(GROUPS).pop(name)
            raise ValueError("Empty agent group")
        if new_group and not agent_roles and not rpc_capabilities and not pubsub_capabilities:
            self.compact_dict.get(GROUPS).pop(name)
            raise ValueError(
                f"agent group {name} should have non empty capabilities. Please pass non empty values "
                "for at least one of the three parameters - agent_roles, rpc_capabilities, pubsub_capabilities")

        group_dict = self.compact_dict.get(GROUPS).get(name)
        # todo validate ids
        current_ids_set = set(group_dict.get('identities', []))
        current_ids_set.update(identities)
        group_dict[IDENTITIES] = list(current_ids_set)

        if agent_roles:
            # todo validate agent_roles
            VolttronAuthzMap.update_agent_roles(group_dict, authz_converter.unstructure(agent_roles))

        VolttronAuthzMap.update_rpc_capabilities(group_dict, authz_converter.unstructure(rpc_capabilities))
        VolttronAuthzMap.update_pubsub_capabilities(group_dict, authz_converter.unstructure(pubsub_capabilities))

        if group_dict.get("agent_roles"):
            self.expand_agent_capabilities(agent_capabilities=self.agent_capabilities,
                                           agent_groups={name: group_dict},
                                           roles=self.compact_dict.get(ROLES))
        else:
            self.expand_agent_capabilities(agent_capabilities=self.agent_capabilities, agent_groups={name: group_dict})
        return True

    def remove_agents_from_group(self, name: str, identities: list[Identity]) -> bool:
        if not self.compact_dict.get(GROUPS) or name not in self.compact_dict[GROUPS]:
            return False
        s = set(self.compact_dict[GROUPS][name][IDENTITIES])
        self.compact_dict[GROUPS][name][IDENTITIES] = list(s - set(identities))
        # expand will only create or merge so reset agent_capabilities to compact_dict value and then expand
        for _id in identities:
            self.agent_capabilities[_id] = copy.deepcopy(self.compact_dict[AGENTS][_id])
        VolttronAuthzMap.expand_agent_capabilities(agent_capabilities=self.agent_capabilities,
                                                   agent_groups={name: self.compact_dict[GROUPS][name]},
                                                   roles=self.compact_dict[ROLES])
        return True

    def add_agents_to_group(self, name: str, identities: list[Identity]):
        if not self.compact_dict.get(GROUPS) or name not in self.compact_dict[GROUPS]:
            return False
        # TODO validate identity
        s = set(self.compact_dict[GROUPS][name][IDENTITIES])
        s.update(identities)
        self.compact_dict[GROUPS][name][IDENTITIES] = list(s)
        VolttronAuthzMap.expand_agent_capabilities(agent_capabilities=self.agent_capabilities,
                                                   agent_groups={name: self.compact_dict[GROUPS][name]},
                                                   roles=self.compact_dict[ROLES])
        return True

    def create_or_merge_agent_authz(self,
                                    *,
                                    identity: str,
                                    protected_rpcs: list[str] = None,
                                    agent_roles: AgentRoles = None,
                                    rpc_capabilities: RPCCapabilities = None,
                                    pubsub_capabilities: PubsubCapabilities = None,
                                    comments: str = None,
                                    **kwargs) -> bool:

        if not identity:
            raise ValueError("agent identity is mandatory")

        if not self.compact_dict.get(AGENTS):
            self.compact_dict[AGENTS] = dict()
            self.agent_capabilities = dict()
        if identity not in self.compact_dict.get(AGENTS):
            self.compact_dict.get(AGENTS)[identity] = dict()

        agent_dict = self.compact_dict.get(AGENTS).get(identity)

        if protected_rpcs:
            current_rpc_set = set(agent_dict.get("protected_rpcs", list()))
            current_rpc_set.update(protected_rpcs)
            agent_dict["protected_rpcs"] = list(current_rpc_set)

        if agent_roles:
            VolttronAuthzMap.update_agent_roles(agent_dict, authz_converter.unstructure(agent_roles))
        if rpc_capabilities:
            VolttronAuthzMap.update_rpc_capabilities(agent_dict, authz_converter.unstructure(rpc_capabilities))
        if pubsub_capabilities:
            VolttronAuthzMap.update_pubsub_capabilities(agent_dict, authz_converter.unstructure(pubsub_capabilities))
        if comments:
            agent_dict["comments"] = comments

        self.agent_capabilities[identity] = copy.deepcopy(agent_dict)
        VolttronAuthzMap.expand_agent_capabilities(agent_capabilities=self.agent_capabilities,
                                                   agent_groups=self.compact_dict.get(GROUPS),
                                                   roles=self.compact_dict.get(ROLES))
        return True

    def get_protected_rpcs(self, identity) -> list[str]:
        id_authz = self.agent_capabilities.get(identity)
        if not id_authz:
            raise ValueError(f"Invalid agent identity {identity}")
        return id_authz.get("protected_rpcs", [])

    def create_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        _topics = self.compact_dict.get("protected_topics", [])
        if not _topics:
            self.compact_dict["protected_topics"] = topic_name_patterns
            return True

        return_value = False
        for topic_name_pattern in topic_name_patterns:
            if topic_name_pattern not in _topics:
                _topics.append(topic_name_pattern)
                return_value = True
        return return_value

    def remove_protected_topics(self, *, topic_name_patterns: list[str]) -> bool:
        _topics = self.compact_dict.get("protected_topics")
        if not _topics:
            return False

        return_value = False
        for topic_name_pattern in topic_name_patterns:
            if topic_name_pattern in _topics:
                _topics.remove(topic_name_pattern)
                return_value = True
        return return_value

    def is_protected_topic(self, *, topic_name_pattern: str) -> bool:
        from volttron.utils import is_regex

        # 1. Check that we have any protected topics.
        if not self.compact_dict.get('protected_topics'):
            return False

        # TODO make the compile a part of the create and merge rather than having it as a loop here.
        # 2. Determine if the topic_name_pattern is a regex or not.  If not then
        #    the string is a topic prefix so we add a .* to the end so that it
        #    becomes a regular expression.
        for tp in self.compact_dict["protected_topics"]:
            if is_regex(tp):
                compiled = re.compile(tp[1:-1])
            else:
                compiled = re.compile(tp + ".*")
            if compiled.match(topic_name_pattern):
                return True

        return False

    def remove_agent_authorization(self, identity: Identity):
        if not self.compact_dict.get(AGENTS) or identity not in self.compact_dict.get(AGENTS):
            return False
        else:
            del self.compact_dict.get(AGENTS)[identity]
            return True

    def remove_agent_group(self, name: str):
        if not self.compact_dict.get(GROUPS) or name not in self.compact_dict.get(GROUPS):
            return False
        else:
            del self.compact_dict.get(GROUPS)[name]
            return True

    def remove_role(self, name: str):
        if not self.compact_dict.get(ROLES) or name not in self.compact_dict.get(ROLES):
            return False
        else:
            del self.compact_dict.get(ROLES)[name]
            return True


def unstructure_authz_map(instance: VolttronAuthzMap):
    authz_map_dict = dict()
    if instance.protected_topics:
        authz_map_dict["protected_topics"] = instance.protected_topics
    if instance.roles:
        authz_map_dict[ROLES] = authz_converter.unstructure(instance.roles)
    if instance.agent_groups:
        authz_map_dict[GROUPS] = authz_converter.unstructure(instance.agent_groups)
    if instance.agents:
        authz_map_dict[AGENTS] = authz_converter.unstructure(instance.agents)
    return authz_map_dict


authz_converter.register_unstructure_hook(RPCCapabilities, unstructure_rpc_capabilities)
authz_converter.register_unstructure_hook(PubsubCapabilities, unstructure_pubsub_capabilities)
authz_converter.register_unstructure_hook(Roles, unstructure_roles)
authz_converter.register_unstructure_hook(AgentGroups, unstructure_agent_groups)
authz_converter.register_unstructure_hook(AgentRoles, unstructure_agent_roles)
authz_converter.register_unstructure_hook(Agents, unstructure_agents)
authz_converter.register_unstructure_hook(VolttronAuthzMap, unstructure_authz_map)

###############
# test
###############
if __name__ == "__main__":
    # test unstructure hooks
    test_r1 = RPCCapability("id.rpc1")
    test_r2 = RPCCapability("id2.rpc2", {"id": "id2", "p2": "v2"})
    test_rpc_list = RPCCapabilities([test_r1, test_r2])
    print(test_rpc_list)
    test_rpc_dict = authz_converter.unstructure(test_rpc_list)
    print(test_rpc_dict)
    test_pubsub_list = PubsubCapabilities([
        PubsubCapability(topic_pattern="device/*", topic_access="pubsub"),
        PubsubCapability(topic_pattern="*", topic_access="pubsub")
    ])
    print(test_pubsub_list)
    pubsub_dict = authz_converter.unstructure(test_pubsub_list)
    print(pubsub_dict)

    test_role = Role("admin", test_rpc_list, test_pubsub_list)
    print(test_role)
    test_role_dict = authz_converter.unstructure(test_role)
    print(test_role_dict)
    test_roles = Roles([Role("admin", test_rpc_list), Role("new_role", pubsub_capabilities=test_pubsub_list)])
    print(test_roles)
    test_roles_dict = authz_converter.unstructure(test_roles)
    print(test_roles_dict)

    test_group = AgentGroups([
        AgentGroup("admin_agents", ["volttron.ctl", "config.store"],
                   agent_roles=AgentRoles([AgentRole(role_name="admin")]),
                   pubsub_capabilities=test_pubsub_list)
    ])
    print(test_group)
    test_group_dict = authz_converter.unstructure(test_group)
    print(test_group_dict)

    test_agents = Agents([
        Agent("volttron.ctl", agent_roles=AgentRoles([AgentRole("admin")])),
    # Agent("hist1", rpc_capabilities=test_rpc_list),
        Agent("listener1",
              agent_roles=AgentRoles([AgentRole(role_name="role1", param_restrictions={"param1": "value1"})]))
    ])
    print(test_agents)
    test_agents_dict = authz_converter.unstructure(test_agents)
    print(test_agents_dict)

    # # Test rpc capabilities update
    # current_rpc_caps = ["string value", {"p1": {"param1": "value1", "param2": "value3", "param3": "value3"}}]
    # element = {"p1": {"param1": "value1", "param2": "value2"}}
    # ## element = {"p3": {1: "12"}}
    # ## element = "sasdf"
    # if isinstance(element, dict):
    #     for i, cur_cap in enumerate(current_rpc_caps):
    #         if isinstance(cur_cap, dict) and next(iter(element)) == next(iter(cur_cap)):
    #             new_param_restrict_value = element[next(iter(element))]
    #             new_param_restrict_value.update(cur_cap[next(iter(cur_cap))])
    #             current_rpc_caps[i] = element
    #             break
    #     else:
    #         current_rpc_caps.append(element)
    # elif element not in current_rpc_caps:
    #     current_rpc_caps.append(element)
    # print(current_rpc_caps)

    # #test expand agent caps



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""
This is the main entry point to the VOLTTRON server. The first thing that
should happen is setting up of logging and verbosity for the server.  After
that we hand off to the run_server method, which will start the server process.
"""
import logging
import logging.config
import os
from pathlib import Path
import sys

import yaml

from volttron.server.logs import get_default_logging_config, log_to_file

file_to_log_to: str | None = None
total_count: int = 0
logging_config: Path | str | None = None

arg_iter = iter(sys.argv)
for arg in arg_iter:
    vcount = 0

    if arg.startswith("-v"):
        vcount = arg.count("v")
    elif arg == "--verbose":
        vcount = 1
    elif arg == "--log" or arg == "-l":
        file_to_log_to = next(arg_iter)

    elif arg == "--log-config" or arg == "-L":
        # Find the index of the log configuration so we can replace the
        # file with absolute path.
        try:
            index = sys.argv.index("--log-config")
        except ValueError:
            index = sys.argv.index("-L")

        # Get the logging config file and verify it's existence.
        logging_config = Path(sys.argv[index + 1])
        if not logging_config.exists():
            sys.stderr.write(f"Invalid --log-config file passed {logging_config}")
            sys.exit(10)

        # Finally set the logging parameter to the absolute path
        sys.argv[index + 1] = logging_config.absolute().as_posix()

    total_count += vcount

# If the user has specified a logging configuration then allow them full access to all the
# responsibility of logging.
if logging_config:
    with open(logging_config, 'rt') as f:
        cfg = yaml.safe_load(f.read())

    logging.config.dictConfig(cfg)

# The user wants to output all to a file so we can do that for them.
if file_to_log_to:
    log_to_file(file_to_log_to, total_count, handler_class=logging.handlers.WatchedFileHandler)

# Normal here just use the default stream handler setup.
else:
    logging.config.dictConfig(get_default_logging_config(level=total_count))

from volttron.server.run_server import _main

# Append PYTHONPATH variables to the system environment
python_path = os.environ.get("PYTHONPATH")
if python_path:
    for pth in python_path.split(":"):
        if pth not in sys.path:
            sys.path.insert(0, pth)

_main()



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
"""VOLTTRON platform™ topic templates.

Templates of standard topics.  Fields in the templates are replaced by
calling the template with the field value included in the keyword
arguments.  Fields are replaced from left to right as long as a
replacement can be made.  Once a field is reached which cannot be
replaced, everything in the replaced portion up to the last double slash
is returned.  Fields cannot be skipped, but may be included
unsubstituted by using None for the field value.  Below are some
examples to demonstrate.

    >>> T = _('root/{top}//{middle}//{bottom}')
    >>> T()
    Topic(u'root')
    >>> T(top='first')
    Topic(u'root/first')
    >>> T(top='first', middle='second')
    Topic(u'root/first/second')
    >>> T(top='first', middle='second', bottom='third')
    Topic(u'root/first/second/third')
    >>> unicode(T(top='first', middle='second', bottom='third'))
    u'root/first/second/third'
    >>> T(top='first', bottom='third')
    ValueError: unused keyword argument: bottom
    >>> T(top='first', middle=None, bottom='third')
    Topic(u'root/first/{middle}/third')
    >>> T(top='first', middle=None, bottom='third')(middle='.')
    Topic(u'root/first/third')
    >>> T(top='first', middle=None, bottom='third')(middle='..')
    Topic(u'root/third')
    >>> T._(top='first', middle=None, bottom='third')
    Topic(u'root/first//{middle}//third')
"""

import os
import platform

from .utils import Topic as _

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"
__copyright__ = "Copyright (c) 2016, Battelle Memorial Institute"
__license__ = "Apache 2.0"

ALERTS_BASE = _("alerts")
ALERTS = _("alerts/{agent_class}/{agent_identity}")    # /{agent_class}/{publickey}/{alert_key}')

HEARTBEAT = _("heartbeats")
PLATFORM_BASE = _("platform")
PLATFORM_SEND_EMAIL = _("platform/send_email")
PLATFORM = _("platform/{subtopic}")
PLATFORM_SHUTDOWN = PLATFORM(subtopic="shutdown")
PLATFORM_VCP_DEVICES = _("platforms/{platform_uuid}/devices/{topic}")

RECORD_BASE = _("record")
RECORD = _("record/{subtopic}")

MARKET_BASE = _("market/{subtopic}")
MARKET_RESERVE = _(MARKET_BASE.replace("{subtopic}", "reserve"))
MARKET_BID = _(MARKET_BASE.replace("{subtopic}", "bid"))
MARKET_CLEAR = _(MARKET_BASE.replace("{subtopic}", "cleared_price"))
MARKET_AGGREGATE = _(MARKET_BASE.replace("{subtopic}", "aggregate"))
MARKET_ERROR = _(MARKET_BASE.replace("{subtopic}", "error"))
MARKET_RECORD = _(RECORD.replace("{subtopic}", MARKET_CLEAR))

AGENT_SHUTDOWN = _("agent/{agent}/shutdown")
AGENT_PING = _("agent/ping/{}/{}/{{cookie}}".format(platform.uname()[1], os.getpid()))

LOGGER_BASE = _("datalogger")
LOGGER = _("datalogger/{subtopic}")
LOGGER_LOG = LOGGER(subtopic="log")
LOGGER_STATUS = LOGGER(subtopic="status")

DRIVER_TOPIC_BASE = "devices"
DRIVER_TOPIC_ALL = "all"
DEVICES_PATH = _("{base}//{node}//{campus}//{building}//{unit}//{path!S}//{point}")
_DEVICES_VALUE = _(DEVICES_PATH.replace("{base}", DRIVER_TOPIC_BASE))
DEVICES_VALUE = _(_DEVICES_VALUE.replace("{node}/", ""))

# For use with RPC calls that require a device path. A plain device path with no prefix.
# Should be used when working with devices via the actuator agent RPC calls:
# get_point, set_point, revert_point, revert_device, and request_new_schedule.
RPC_DEVICE_PATH = _(DEVICES_PATH.replace("{base}//{node}//", ""))

ANALYSIS_PATH = _("{base}//{analysis_name}//{campus}//{building}//{unit}//{point}")
ANALYSIS_TOPIC_BASE = "analysis"
ANALYSIS_VALUE = _(ANALYSIS_PATH.replace("{base}", ANALYSIS_TOPIC_BASE))

ACTUATOR_GET = _(_DEVICES_VALUE.replace("{node}", "actuators/get"))
ACTUATOR_SET = _(_DEVICES_VALUE.replace("{node}", "actuators/set"))
ACTUATOR_REVERT_POINT = _(_DEVICES_VALUE.replace("{node}", "actuators/revert/point"))
ACTUATOR_REVERT_DEVICE = _(_DEVICES_VALUE.replace("{node}", "actuators/revert/device"))

_ACTUATOR_SCHEDULE = _(("{base}/actuators/schedule/{op}").replace("{base}", DRIVER_TOPIC_BASE))
ACTUATOR_SCHEDULE_REQUEST = _(_ACTUATOR_SCHEDULE.replace("{op}", "request"))
ACTUATOR_SCHEDULE_RESULT = _(_ACTUATOR_SCHEDULE.replace("{op}", "result"))
ACTUATOR_SCHEDULE_ANNOUNCE_RAW = _(_ACTUATOR_SCHEDULE.replace("{op}", "announce/{device}"))

# This is a convenience topic for agent listening for announcements
# and want to use the {campus}//{building}//{unit} style replacement
ACTUATOR_SCHEDULE_ANNOUNCE = _(
    ACTUATOR_SCHEDULE_ANNOUNCE_RAW.replace("{device}", "{campus}//{building}//{unit}"))

# Added by CHA to be used as the root of all actuators for working within
# base_historian.py.
ACTUATOR_BASE = _("actuators")
ACTUATOR = _(_DEVICES_VALUE.replace("{node}", ACTUATOR_BASE))
ACTUATOR_ERROR = _(_DEVICES_VALUE.replace("{node}", "actuators/error"))
ACTUATOR_VALUE = _(_DEVICES_VALUE.replace("{node}", "actuators/value"))

# Ragardless of the interface used (RPC vs pubsub) when an agent
# attempts to set a point it is announced on this topic.
# This is intended to inable a historian to capture all attempted writes.
ACTUATOR_WRITE = _(_DEVICES_VALUE.replace("{node}", "actuators/write"))
ACTUATOR_REVERTED_POINT = _(_DEVICES_VALUE.replace("{node}", "actuators/reverted/point"))
ACTUATOR_REVERTED_DEVICE = _(_DEVICES_VALUE.replace("{node}", "actuators/reverted/device"))

BASE_ARCHIVER_REQUEST = _("archiver/request")
BASE_ARCHIVER_FULL_REQUEST = _("archiver/full/request")
BASE_ARCHIVER_RESPONSE = _("archiver/response")

_ARCHIVER = _("{base}/{campus}//{building}//{unit}//{point}")
_ARCHIVER_UNIT = _("{base}/{campus}//{building}//{unit}")
ARCHIVER_REQUEST = _(_ARCHIVER.replace("{base}", BASE_ARCHIVER_REQUEST))
ARCHIVER_RESPONSE = _(_ARCHIVER.replace("{base}", BASE_ARCHIVER_RESPONSE))

ARCHIVER_FULL_UNIT_REQUEST = _(_ARCHIVER_UNIT.replace("{base}", BASE_ARCHIVER_FULL_REQUEST))

OPENADR_STATUS = _("openadr/status")
OPENADR_EVENT = _("openadr/event")

_SUBSCRIPTIONS = _("subscriptions/{op}/{{topic}}")
SUBSCRIPTIONS_LIST = _(_SUBSCRIPTIONS.format(op="list"))
SUBSCRIPTIONS_ADD = _(_SUBSCRIPTIONS.format(op="add"))
SUBSCRIPTIONS_REMOVE = _(_SUBSCRIPTIONS.format(op="remove"))

_BUILDING = _("building/{op}/{{campus}}//{{building}}//{{topic}}")
BUILDING_SEND = _(_BUILDING.format(op="send"))
BUILDING_RECV = _(_BUILDING.format(op="recv"))
BUILDING_ERROR = _(_BUILDING.format(op="error"))

CONFIG_TOPIC_BASE = "config"
CONFIG_PATH = _("{base}//{action}//{category}//{name}")
_CONFIG_VALUE = _(CONFIG_PATH.replace("{base}", CONFIG_TOPIC_BASE))
CONFIG_ADD = _(_CONFIG_VALUE.replace("{action}", "add"))
CONFIG_REMOVE = _(_CONFIG_VALUE.replace("{action}", "remove"))
CONFIG_UPDATE = _(_CONFIG_VALUE.replace("{action}", "update"))

DRIVER_CONFIG_ADD = _(_CONFIG_VALUE.replace("{category}", "driver"))
DRIVER_CONFIG_REMOVE = _(_CONFIG_VALUE.replace("{category}", "driver"))
DRIVER_CONFIG_UPDATE = _(_CONFIG_VALUE.replace("{category}", "driver"))

WEATHER_BASE = "weather"
WEATHER_REQUEST = "weather/request"

WEATHER2_BASE = "weather2"
WEATHER2_REQUEST = "weather2/request"
WEATHER2_POLL = "weather2/polling"
WEATHER2_RESPONSE = "weather2/response"
WEATHER2_ERROR = "weather2/error"

BACNET_INFO_BASE = _("bacnet/{indication}")
BACNET_I_AM = _(BACNET_INFO_BASE.replace("{indication}", "i_am"))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from dataclass_wizard import JSONSerializable


class Message(object):
    """Message object returned form Socket.recv_vip_object()."""

    def __init__(self, **kwargs):
        self.__dict__ = kwargs

    def __repr__(self):
        attrs = ", ".join("%r: %r" % (
            name,
            [x for x in value] if isinstance(value, (list, tuple)) else value,
        ) for name, value in self.__dict__.items())
        return "%s(**{%s})" % (self.__class__.__name__, attrs)

    @staticmethod
    def create_message(*,
                       peer: str,
                       user: str,
                       subsystem: str,
                       msg_id: str,
                       args: list = []) -> Message:
        return Message(peer=peer, subsystem=subsystem, msg_id=msg_id, user=user, args=args)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

__all__ = ["create_file_if_missing"]

import os
import errno
import logging

_log = logging.getLogger(__name__)


def create_file_if_missing(path, permission=0o660, contents=None):
    dirname = os.path.dirname(path)
    if dirname and not os.path.exists(dirname):
        try:
            os.makedirs(dirname)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    try:
        with open(path) as fd:
            pass
    except IOError as exc:
        if exc.errno != errno.ENOENT:
            raise
        _log.debug("missing file %s", path)
        _log.info("creating file %s", path)
        fd = os.open(path, os.O_CREAT | os.O_WRONLY, permission)
        success = False
        try:
            if contents:
                contents = (contents if isinstance(contents, bytes) else contents.encode("utf-8"))
                os.write(fd, contents)
                success = True
        except Exception as e:
            raise e
        finally:
            os.close(fd)
        return success



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import os
from configparser import ConfigParser

from volttron.utils import ClientContext as cc

_log = logging.getLogger(__name__)


def store_message_bus_config(message_bus, instance_name):
    # If there is no config file or home directory yet, create volttron_home
    # and config file
    if not instance_name:
        raise ValueError("Instance name should be a valid string and should "
                         "be unique within a network of volttron instances "
                         "that communicate with each other. start volttron "
                         "process with '--instance-name <your instance>' if "
                         "you are running this instance for the first time. "
                         "Or add instance-name = <instance name> in "
                         "vhome/config")

    v_home = cc.get_volttron_home()
    config_path = os.path.join(v_home, "config")
    if os.path.exists(config_path):
        config = ConfigParser()
        config.read(config_path)
        config.set("volttron", "messagebus", message_bus)
        config.set("volttron", "instance-name", instance_name)
        with open(config_path, "w") as configfile:
            config.write(configfile)
    else:
        if not os.path.exists(v_home):
            os.makedirs(v_home, 0o755)
        config = ConfigParser()
        config.add_section("volttron")
        config.set("volttron", "messagebus", message_bus)
        config.set("volttron", "instance-name", instance_name)

        with open(config_path, "w") as configfile:
            config.write(configfile)
        # all agents need read access to config file
        os.chmod(config_path, 0o744)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
"""VOLTTRON platform™ messaging header name constants."""

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"
__copyright__ = "Copyright (c) 2016, Battelle Memorial Institute"
__license__ = "Apache 2.0"

CONTENT_TYPE = type("ContentTypeStr", (str, ), {
    "JSON": "application/json",
    "PLAIN_TEXT": "text/plain"
})("Content-Type")

DATE = "Date"

TIMESTAMP = "TimeStamp"

SYNC_TIMESTAMP = "SynchronizedTimeStamp"

FROM = "From"
TO = "To"

REQUESTER_ID = "requesterID"
COOKIE = "Cookie"


class Headers(dict):
    """Case-insensitive dictionary for HTTP-like headers."""

    class Key(str):

        def __new__(cls, value):
            string = str(value)
            obj = str.__new__(cls, string.lower())
            obj._orig = string
            return obj

        def __str__(self):
            return self._orig

        def __repr__(self):
            return repr(self._orig)

    def __init__(self, *args, **kwargs):
        Key = self.__class__.Key
        obj = super(Headers, self).__init__(
            ((Key(key), value) for key, value in dict(*args, **kwargs).items()))

    def __contains__(self, key):
        return super(Headers, self).__contains__(str(key).lower())

    def get(self, key, default=None):
        return super(Headers, self).get(str(key).lower(), default)

    def __getitem__(self, key):
        return super(Headers, self).__getitem__(str(key).lower())

    def __setitem__(self, key, value):
        super(Headers, self).__setitem__(self.__class__.Key(key), value)

    def __delitem__(self, key):
        super(Headers, self).__delitem__(str(key).lower())

    def copy(self):
        return Headers(super(Headers, self).copy())

    @property
    def dict(self):
        """Return a dictionary with originally-cased keys."""
        return {str(k): v for k, v in self.items()}

    def setdefault(self, key, value):
        return super(Headers, self).setdefault(self.__class__.Key(key), value)

    def update(self, *args, **kwargs):
        Key = self.__class__.Key
        obj = super(Headers, self).update(
            ((Key(key), value) for key, value in dict(*args, **kwargs).items()))

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, super(Headers, self).__repr__())



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2023 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import gevent

from volttron.client.vip.agent.errors import Unreachable
from volttron.utils.jsonrpc import MethodNotFound


def print_rpc_list(peers, code=False):
    for peer in peers:
        print(f"{peer}")
        for method in peers[peer]:
            if code:
                print(f"\tself.vip.rpc.call({peer}, {method}).get()")
            else:
                print(f"\t{method}")


def print_rpc_methods(opts, peer_method_metadata, code=False):
    for peer in peer_method_metadata:
        if code is True:
            pass
        else:
            print(f"{peer}")
        for method in peer_method_metadata[peer]:
            params = peer_method_metadata[peer][method].get("params",
                                                            "No parameters for this method.")
            if code is True:
                if len(params) == 0:
                    print(f"self.vip.rpc.call({peer}, {method}).get()")
                else:
                    print(f"self.vip.rpc.call({peer}, {method}, "
                          f"{[param for param in params]}).get()")
                continue
            else:
                print(f"\t{method}")
                if opts.verbose == True:
                    print("\tDocumentation:")
                    doc = (peer_method_metadata[peer][method].get(
                        "doc", "No documentation for this method.").replace("\n", "\n\t\t"))
                    print(f"\t\t{doc}\n")
            print("\tParameters:")
            if type(params) is str:
                print(f"\t\t{params}")
            else:
                for param in params:
                    print(f"\t\t{param}:\n\t\t\t{params[param]}")


def list_agents_rpc(opts):
    conn = opts.connection
    try:
        peers = sorted(conn.call("peerlist"))
    except Exception as e:
        print(e)
    if opts.by_vip == True or len(opts.pattern) == 1:
        peers = [peer for peer in peers if peer in opts.pattern]
    elif len(opts.pattern) > 1:
        peer = opts.pattern[0]
        methods = opts.pattern[1:]
        peer_method_metadata = {peer: {}}
        for method in methods:
            try:
                peer_method_metadata[peer][method] = conn.server.vip.rpc.call(
                    peer, f"{method}.inspect").get(timeout=4)
                authorized_capabilities = conn.server.vip.rpc.call(peer,
                                                                   "auth.get_rpc_authorizations",
                                                                   method).get(timeout=4)
                peer_method_metadata[peer][method][
                    "authorized_capabilities"] = f"Authorized capabilities: {authorized_capabilities}"
            except gevent.Timeout:
                print(f"{peer} has timed out.")
            except Unreachable:
                print(f"{peer} is unreachable")
            except MethodNotFound as e:
                print(e)

        # _stdout.write(f"{peer_method_metadata}\n")
        print_rpc_methods(opts, peer_method_metadata)
        return
    peer_methods = {}
    for peer in peers:
        try:
            peer_methods[peer] = conn.server.vip.rpc.call(peer,
                                                          "inspect").get(timeout=4)["methods"]
        except gevent.Timeout:
            print(f"{peer} has timed out")
        except Unreachable:
            print(f"{peer} is unreachable")
        except MethodNotFound as e:
            print(e)

    if opts.verbose is True:
        print_rpc_list(peer_methods)
        # for peer in peer_methods:
        #     _stdout.write(f"{peer}:{peer_methods[peer]}\n")
    else:
        for peer in peer_methods:
            peer_methods[peer] = [method for method in peer_methods[peer] if "." not in method]
            # _stdout.write(f"{peer}:{peer_methods[peer]}\n")
        print_rpc_list(peer_methods)


def list_agent_rpc_code(opts):
    conn = opts.connection
    try:
        peers = sorted(conn.call("peerlist"))
    except Exception as e:
        print(e)
    if len(opts.pattern) == 1:
        peers = [peer for peer in peers if peer in opts.pattern]
    elif len(opts.pattern) > 1:
        peer = opts.pattern[0]
        methods = opts.pattern[1:]
        peer_method_metadata = {peer: {}}
        for method in methods:
            try:
                peer_method_metadata[peer][method] = conn.server.vip.rpc.call(
                    peer, f"{method}.inspect").get(timeout=4)
            except gevent.Timeout:
                print(f"{peer} has timed out.")
            except Unreachable:
                print(f"{peer} is unreachable")
            except MethodNotFound as e:
                print(e)

        # _stdout.write(f"{peer_method_metadata}\n")
        print_rpc_methods(opts, peer_method_metadata, code=True)
        return

    peer_methods = {}
    for peer in peers:
        try:
            peer_methods[peer] = conn.server.vip.rpc.call(peer,
                                                          "inspect").get(timeout=4)["methods"]
        except gevent.Timeout:
            print(f"{peer} has timed out.")
        except Unreachable:
            print(f"{peer} is unreachable")
        except MethodNotFound as e:
            print(e)

    if opts.verbose is True:
        pass
    else:
        for peer in peer_methods:
            peer_methods[peer] = [method for method in peer_methods[peer] if "." not in method]

    peer_method_metadata = {}
    for peer in peer_methods:
        peer_method_metadata[peer] = {}
        for method in peer_methods[peer]:
            try:
                peer_method_metadata[peer][method] = conn.server.vip.rpc.call(
                    peer, f"{method}.inspect").get(timeout=4)
            except gevent.Timeout:
                print(f"{peer} has timed out")
            except Unreachable:
                print(f"{peer} is unreachable")
            except MethodNotFound as e:
                print(e)
    print_rpc_methods(opts, peer_method_metadata, code=True)


def add_rpc_agent_parser(add_parser_fn):
    rpc_ctl = add_parser_fn("rpc", help="rpc controls")

    rpc_subparsers = rpc_ctl.add_subparsers(title="subcommands", metavar="", dest="store_commands")

    rpc_code = add_parser_fn(
        "code",
        subparser=rpc_subparsers,
        help="shows how to use rpc call in other agents",
    )

    rpc_code.add_argument("pattern", nargs="*", help="Identity of agent, followed by method(s)"
                          "")
    rpc_code.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="list all subsystem rpc methods in addition to the agent's rpc "
        "methods",
    )

    rpc_code.set_defaults(func=list_agent_rpc_code, min_uuid_len=1)

    rpc_list = add_parser_fn("list",
                             subparser=rpc_subparsers,
                             help="lists all agents and their rpc methods")

    rpc_list.add_argument("-i",
                          "--vip",
                          dest="by_vip",
                          action="store_true",
                          help="filter by vip identity")

    rpc_list.add_argument("pattern", nargs="*", help="UUID or name of agent")

    rpc_list.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="list all subsystem rpc methods in addition to the agent's rpc "
        "methods. If a method "
        "is specified, display the doc-string associated with the "
        "method.",
    )

    rpc_list.set_defaults(func=list_agents_rpc, min_uuid_len=1)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
""" Core package."""
from typing import List

from gevent import monkey

# At this point these are the only things that need to be patched
# and the server and client are working harmoniously with this.
patches = [
    ('ssl', monkey.patch_ssl),
    ('socket', monkey.patch_socket),
    ('os', monkey.patch_os),
]

# patch modules if necessary.  Only if the module hasn't been patched before.
# this could happen if the server code uses the client (which it does).
for module, fn in patches:
    if not monkey.is_module_patched(module):
        fn()

import logging
from urllib.parse import urlparse

from volttron.client.logs import get_logger
from volttron.client.vip.agent.core import Core
from volttron.client.vip.agent.subsystems.auth import Auth
from volttron.client.vip.agent.subsystems.configstore import ConfigStore
from volttron.client.vip.agent.subsystems.health import Health
from volttron.client.vip.agent.subsystems.heartbeat import Heartbeat
from volttron.client.vip.agent.subsystems.hello import Hello
from volttron.client.vip.agent.subsystems.peerlist import PeerList
from volttron.client.vip.agent.subsystems.ping import Ping
from volttron.client.vip.agent.subsystems.pubsub import PubSub
from volttron.client.vip.agent.subsystems.query import Query
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent import Agent
from volttron.types import AbstractAgent
from volttron.client.logs import setup_logging

__all__: List[str] = [
    "Agent", "AbstractAgent", "Core", "RPC", "Hello", "PeerList", "Ping", "PubSub", "Heartbeat", "Health",
    "ConfigStore", "Auth", "Query", "setup_logging"
]

_log = get_logger()
